try:
    from django.urls import include, re_path
except ImportError:
    from django.conf.urls import include, url as re_path

from django.contrib.auth.decorators import login_required
from . import views

urlpatterns = [
    re_path(r"^$", login_required(views.UserPreferenceFormView.as_view()), name="user"),
    re_path(
        r"^(?P<section>[\w\ ]+)$",
        login_required(views.UserPreferenceFormView.as_view()),
        name="user.section",
    ),
]



from django.core.management.base import BaseCommand
from dynamic_preferences.exceptions import NotFoundInRegistry
from dynamic_preferences.models import GlobalPreferenceModel
from dynamic_preferences.registries import (
    global_preferences_registry,
    preference_models,
)
from dynamic_preferences.settings import preferences_settings


def delete_preferences(queryset):
    """
    Delete preferences objects if they are not present in registry.
    Return a list of deleted objects
    """
    deleted = []

    # Iterate through preferences. If an error is raised when accessing
    # preference object, just delete it
    for p in queryset:
        try:
            p.registry.get(section=p.section, name=p.name, fallback=False)
        except NotFoundInRegistry:
            p.delete()
            deleted.append(p)

    return deleted


class Command(BaseCommand):
    help = (
        "Find and delete preferences from database if they don't exist in "
        "registries. Create preferences that are not present in database"
        "(except when invoked with --skip_create)."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--skip_create",
            action="store_true",
            help="Forces to skip the creation step for missing preferences",
        )

    def handle(self, *args, **options):
        skip_create = options["skip_create"]

        # Create needed preferences
        # Global
        if not skip_create:
            self.stdout.write("Creating missing global preferences...")
            manager = global_preferences_registry.manager()
            manager.all()

        deleted = delete_preferences(GlobalPreferenceModel.objects.all())
        message = "Deleted {deleted} global preferences".format(deleted=len(deleted))
        self.stdout.write(message)

        for preference_model, registry in preference_models.items():
            deleted = delete_preferences(preference_model.objects.all())
            message = "Deleted {deleted} {model} preferences".format(
                deleted=len(deleted),
                model=preference_model.__name__,
            )
            self.stdout.write(message)
            if not hasattr(preference_model, "get_instance_model"):
                continue

            if skip_create:
                continue

            message = "Creating missing preferences for {model} model...".format(
                model=preference_model.get_instance_model().__name__,
            )
            self.stdout.write(message)
            for instance in preference_model.get_instance_model().objects.all():
                getattr(instance, preferences_settings.MANAGER_ATTRIBUTE).all()



try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from .settings import preferences_settings
from .exceptions import CachedValueNotFound, DoesNotExist
from .signals import preference_updated


class PreferencesManager(Mapping):

    """Handle retrieving / caching of preferences"""

    def __init__(self, model, registry, **kwargs):
        self.model = model
        self.registry = registry
        self.instance = kwargs.get("instance")

    @property
    def queryset(self):
        qs = self.model.objects.all()
        if self.instance:
            qs = qs.filter(instance=self.instance)
        return qs

    @property
    def cache(self):
        from django.core.cache import caches

        return caches[preferences_settings.CACHE_NAME]

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        section, name = self.parse_lookup(key)
        preference = self.registry.get(section=section, name=name, fallback=False)
        preference.validate(value)
        self.update_db_pref(section=section, name=name, value=value)

    def __repr__(self):
        return repr(self.all())

    def __iter__(self):
        return self.all().__iter__()

    def __len__(self):
        return len(self.all())

    def by_name(self):
        """Return a dictionary with preferences identifiers and values, but without the section name in the identifier"""
        return {
            key.split(preferences_settings.SECTION_KEY_SEPARATOR)[-1]: value
            for key, value in self.all().items()
        }

    def get_by_name(self, name):
        return self.get(self.registry.get_by_name(name).identifier())

    def get_cache_key(self, section, name):
        """Return the cache key corresponding to a given preference"""
        if not self.instance:
            return "dynamic_preferences_{0}_{1}_{2}".format(
                self.model.__name__, section, name
            )
        return "dynamic_preferences_{0}_{1}_{2}_{3}".format(
            self.model.__name__, self.instance.pk, section, name, self.instance.pk
        )

    def from_cache(self, section, name):
        """Return a preference raw_value from cache"""
        cached_value = self.cache.get(
            self.get_cache_key(section, name), CachedValueNotFound
        )

        if cached_value is CachedValueNotFound:
            raise CachedValueNotFound

        if cached_value == preferences_settings.CACHE_NONE_VALUE:
            cached_value = None
        return self.registry.get(section=section, name=name).serializer.deserialize(
            cached_value
        )

    def many_from_cache(self, preferences):
        """
        Return cached value for given preferences
        missing preferences will be skipped
        """
        keys = {p: self.get_cache_key(p.section.name, p.name) for p in preferences}
        cached = self.cache.get_many(list(keys.values()))

        for k, v in cached.items():
            # we replace dummy cached values by None here, if needed
            if v == preferences_settings.CACHE_NONE_VALUE:
                cached[k] = None

        # we have to remap returned value since the underlying cached keys
        # are not usable for an end user
        return {
            p.identifier(): p.serializer.deserialize(cached[k])
            for p, k in keys.items()
            if k in cached
        }

    def to_cache(self, pref):
        """
        Update/create the cache value for the given preference model instance
        """
        key = self.get_cache_key(pref.section, pref.name)
        value = pref.raw_value
        if value is None or value == "":
            # some cache backends refuse to cache None or empty values
            # resulting in more DB queries, so we cache an arbitrary value
            # to ensure the cache is hot (even with empty values)
            value = preferences_settings.CACHE_NONE_VALUE
        self.cache.set(key, value)

    def pref_obj(self, section, name):
        return self.registry.get(section=section, name=name)

    def parse_lookup(self, lookup):
        try:
            section, name = lookup.split(preferences_settings.SECTION_KEY_SEPARATOR)
        except ValueError:
            name = lookup
            section = None
        return section, name

    def get(self, key, no_cache=False):
        """Return the value of a single preference using a dotted path key
        :arg no_cache: if true, the cache is bypassed
        """
        section, name = self.parse_lookup(key)
        preference = self.registry.get(section=section, name=name, fallback=False)
        if no_cache or not preferences_settings.ENABLE_CACHE:
            return self.get_db_pref(section=section, name=name).value

        try:
            return self.from_cache(section, name)
        except CachedValueNotFound:
            pass

        db_pref = self.get_db_pref(section=section, name=name)
        self.to_cache(db_pref)
        return db_pref.value

    def get_db_pref(self, section, name):
        try:
            pref = self.queryset.get(section=section, name=name)
        except self.model.DoesNotExist:
            pref_obj = self.pref_obj(section=section, name=name)
            pref = self.create_db_pref(
                section=section, name=name, value=pref_obj.get("default")
            )

        return pref

    def update_db_pref(self, section, name, value):
        try:
            db_pref = self.queryset.get(section=section, name=name)
            old_value = db_pref.value
            db_pref.value = value
            db_pref.save()
            preference_updated.send(
                sender=self.__class__,
                section=section,
                name=name,
                old_value=old_value,
                new_value=value,
            )
        except self.model.DoesNotExist:
            return self.create_db_pref(section, name, value)

        return db_pref

    def create_db_pref(self, section, name, value):
        kwargs = {
            "section": section,
            "name": name,
        }
        if self.instance:
            kwargs["instance"] = self.instance

        # this is a just a shortcut to get the raw, serialized value
        # so we can pass it to get_or_create
        m = self.model(**kwargs)
        m.value = value
        raw_value = m.raw_value

        db_pref, created = self.model.objects.get_or_create(**kwargs)
        if created and db_pref.raw_value != raw_value:
            db_pref.raw_value = raw_value
            db_pref.save()

        return db_pref

    def all(self):
        """Return a dictionary containing all preferences by section
        Loaded from cache or from db in case of cold cache
        """
        if not preferences_settings.ENABLE_CACHE:
            return self.load_from_db()

        preferences = self.registry.preferences()

        # first we hit the cache once for all existing preferences
        a = self.many_from_cache(preferences)
        if len(a) == len(preferences):
            return a  # avoid database hit if not necessary

        # then we fill those that miss, but exist in the database
        # (just hit the database for all of them, filtering is complicated, and
        # in most cases you'd need to grab the majority of them anyway)
        a.update(self.load_from_db(cache=True))
        return a

    def load_from_db(self, cache=False):
        """Return a dictionary of preferences by section directly from DB"""
        a = {}
        db_prefs = {p.preference.identifier(): p for p in self.queryset}
        for preference in self.registry.preferences():
            try:
                db_pref = db_prefs[preference.identifier()]
            except KeyError:
                db_pref = self.create_db_pref(
                    section=preference.section.name,
                    name=preference.name,
                    value=preference.get("default"),
                )
            else:
                # cache if create_db_pref() hasn't already done so
                if cache:
                    self.to_cache(db_pref)

            a[preference.identifier()] = db_pref.value

        return a



# Generated by Django 3.1 on 2020-08-21 08:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("dynamic_preferences_users", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userpreferencemodel",
            name="name",
            field=models.CharField(db_index=True, max_length=150, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="userpreferencemodel",
            name="raw_value",
            field=models.TextField(blank=True, null=True, verbose_name="Raw Value"),
        ),
        migrations.AlterField(
            model_name="userpreferencemodel",
            name="section",
            field=models.CharField(
                blank=True,
                db_index=True,
                default=None,
                max_length=150,
                null=True,
                verbose_name="Section Name",
            ),
        ),
    ]



from ..registries import PerInstancePreferenceRegistry


class UserPreferenceRegistry(PerInstancePreferenceRegistry):
    section_url_namespace = "dynamic_preferences:user.section"


user_preferences_registry = UserPreferenceRegistry()



class DynamicPreferencesException(Exception):
    detail_default = "An exception occurred with django-dynamic-preferences"

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = str(detail)
        else:
            self.detail = str(self.detail_default)

    def __str__(self):
        return self.detail


class MissingDefault(DynamicPreferencesException):
    detail_default = "You must provide a default value for all preferences"


class NotFoundInRegistry(DynamicPreferencesException, KeyError):
    detail_default = "Preference with this name/section not found in registry"


class DoesNotExist(DynamicPreferencesException):
    detail_default = "Cannot retrieve preference value, ensure the preference is correctly registered and database is synced"


class CachedValueNotFound(DynamicPreferencesException):
    detail_default = "Cached value not found"


class MissingModel(DynamicPreferencesException):
    detail_default = 'You must define a model choice through "model" \
                      or "queryset" attribute'



"""
You'll find here the final, concrete classes of preferences you can use
in your own project.

"""
from django import forms
from django.db.models.signals import pre_delete

from django.core.files.storage import default_storage

from .preferences import AbstractPreference, Section
from .exceptions import MissingModel
from dynamic_preferences.serializers import *
from dynamic_preferences.settings import preferences_settings


class BasePreferenceType(AbstractPreference):
    """
    Used as a base for all other preference classes. You should subclass
    this one if you want to implement your own preference.
    """

    field_class = None
    """
    A form field that will be used to display and edit the preference
    use a class, not an instance.

    :Example:

    .. code-block:: python

        from django import forms

        class MyPreferenceType(BasePreferenceType):
            field_class = forms.CharField
    """

    #: A serializer class (see dynamic_preferences.serializers)
    serializer = None

    field_kwargs = {}
    """
    Additional kwargs to be passed to the form field.

    :Example:

    .. code-block:: python

        class MyPreference(StringPreference):

            field_kwargs = {
                'required': False,
                'initial': 'Hello there'
            }
    """

    @property
    def initial(self):
        return self.get_initial()

    def get_initial(self):
        """
        :return:
            initial data for form field
            from field_attribute['initial'] or default
        """
        return self.field_kwargs.get("initial", self.get("default"))

    @property
    def field(self):
        """
        :return:
            an instance of a form field for this preference, with
            the correct configuration (widget, initial value, validators...)
        """
        return self.setup_field()

    def setup_field(self, **kwargs):
        field_class = self.get("field_class")
        field_kwargs = self.get_field_kwargs()
        field_kwargs.update(kwargs)
        return field_class(**field_kwargs)

    def get_field_kwargs(self):
        """
        Return a dict of arguments to use as parameters for the field
        class instianciation.

        This will use :py:attr:`field_kwargs` as a starter,
        and use sensible defaults for a few attributes:

        - :py:attr:`instance.verbose_name` for the field label
        - :py:attr:`instance.help_text` for the field help text
        - :py:attr:`instance.widget` for the field widget
        - :py:attr:`instance.required` defined if the value is required or not
        - :py:attr:`instance.initial` defined if the initial value
        """
        kwargs = self.field_kwargs.copy()
        kwargs.setdefault("label", self.get("verbose_name"))
        kwargs.setdefault("help_text", self.get("help_text"))
        kwargs.setdefault("widget", self.get("widget"))
        kwargs.setdefault("required", self.get("required"))
        kwargs.setdefault("initial", self.initial)
        kwargs.setdefault("validators", [])
        kwargs["validators"].append(self.validate)
        return kwargs

    def api_repr(self, value):
        """
        Used only to represent a preference value using Rest Framework
        """
        return value

    def get_api_additional_data(self):
        """
        Additional data to serialize for use on front-end side, for example
        """
        return {}

    def get_api_field_data(self):
        """
        Field data to serialize for use on front-end side, for example
        will include choices available for a choice field
        """
        field = self.setup_field()
        d = {
            "class": field.__class__.__name__,
            "widget": {"class": field.widget.__class__.__name__},
        }

        try:
            d["input_type"] = field.widget.input_type
        except AttributeError:
            # some widgets, such as Select do not have an input type
            # in django < 1.11
            d["input_type"] = None

        return d

    def validate(self, value):
        """
        Used to implement custom cleaning logic for use in forms
        and serializers. The method will be passed as a validator to
        the preference form field.

        :Example:

        .. code-block:: python

            def validate(self, value):
                if value == '42':
                    raise ValidationError('Wrong value!')
        """
        return


class BooleanPreference(BasePreferenceType):
    """
    A preference type that stores a boolean.
    """

    field_class = forms.BooleanField
    serializer = BooleanSerializer
    required = False


class IntegerPreference(BasePreferenceType):
    """
    A preference type that stores an integer.
    """

    field_class = forms.IntegerField
    serializer = IntegerSerializer


IntPreference = IntegerPreference


class DecimalPreference(BasePreferenceType):
    """
    A preference type that stores a :py:class:`decimal.Decimal`.
    """

    field_class = forms.DecimalField
    serializer = DecimalSerializer


class FloatPreference(BasePreferenceType):
    """
    A preference type that stores a float.
    """

    field_class = forms.FloatField
    serializer = FloatSerializer


class StringPreference(BasePreferenceType):
    """
    A preference type that stores a string.
    """

    field_class = forms.CharField
    serializer = StringSerializer


class LongStringPreference(StringPreference):
    """
    A preference type that stores a string, but with a textarea widget.
    """

    widget = forms.Textarea


class ChoicePreference(BasePreferenceType):
    """
    A preference type that stores a string among a list of choices.
    """

    choices = ()
    """
    Expects the same values as for django :py:class:`forms.ChoiceField`.

    :Example:

    .. code-block:: python

        class MyChoicePreference(ChoicePreference):
            choices = [
                ('c', 'Carrot'),
                ('t', 'Tomato'),
            ]
    """
    field_class = forms.ChoiceField
    serializer = StringSerializer

    def get_field_kwargs(self):
        field_kwargs = super(ChoicePreference, self).get_field_kwargs()
        field_kwargs["choices"] = self.get("choices") or self.field_attribute["initial"]
        return field_kwargs

    def get_api_additional_data(self):
        d = super(ChoicePreference, self).get_api_additional_data()
        d["choices"] = self.get("choices")
        return d

    def get_choice_values(self):
        return [c[0] for c in self.get("choices")]

    def validate(self, value):
        if value not in self.get_choice_values():
            raise forms.ValidationError("{} is not a valid choice".format(value))


def create_deletion_handler(preference):
    """
    Will generate a dynamic handler to purge related preference
    on instance deletion
    """

    def delete_related_preferences(sender, instance, *args, **kwargs):
        queryset = preference.registry.preference_model.objects.filter(
            name=preference.name, section=preference.section
        )
        related_preferences = queryset.filter(
            raw_value=preference.serializer.serialize(instance)
        )
        related_preferences.delete()

    return delete_related_preferences


class ModelChoicePreference(BasePreferenceType):
    """
    A preference type that stores a reference to a model instance.

    :Example:

    .. code-block:: python

        from myapp.blog.models import BlogEntry

        @registry.register
        class FeaturedEntry(ModelChoicePreference):
            section = Section('blog')
            name = 'featured_entry'
            queryset = BlogEntry.objects.filter(status='published')

        blog_entry = BlogEntry.objects.get(pk=12)
        manager['blog__featured_entry'] = blog_entry

        # accessing the value will return the model instance
        assert manager['blog__featured_entry'].pk == 12

    .. note::

        You should provide either the :py:attr:`queryset` or :py:attr:`model`
        attribute
    """

    field_class = forms.ModelChoiceField
    serializer_class = ModelSerializer

    model = None
    """
    Which model class to link the preference to. You can skip this if you
    define the :py:attr:`queryset` attribute.
    """

    queryset = None
    """
    A queryset to filter available model instances.
    """
    signals_handlers = {}

    def __init__(self, *args, **kwargs):
        super(ModelChoicePreference, self).__init__(*args, **kwargs)

        if self.model is not None:
            # Set queryset following model attribute
            self.queryset = self.model.objects.all()
        elif self.queryset is not None:
            # Set model following queryset attribute
            self.model = self.queryset.model
        else:
            raise MissingModel

        self.serializer = self.serializer_class(self.model)

        self._setup_signals()

    def _setup_signals(self):
        handler = create_deletion_handler(self)
        # We need to keep a reference to the handler or it will cause
        # weakref to die and our handler will not be called
        self.signals_handlers["pre_delete"] = [handler]
        pre_delete.connect(handler, sender=self.model)

    def get_field_kwargs(self):
        kw = super(ModelChoicePreference, self).get_field_kwargs()
        kw["queryset"] = self.get("queryset")
        return kw

    def api_repr(self, value):
        if not value:
            return None
        if value.__class__.__name__ == "QuerySet":
            return [val.pk for val in value]
        return value.pk


class ModelMultipleChoicePreference(ModelChoicePreference):
    """
    A preference type that stores a reference list to the model instances.

    :Example:

    .. code-block:: python

        from myapp.blog.models import BlogEntry

        @registry.register
        class FeaturedEntries(ModelMultipleChoicePreference):
            section = Section('blog')
            name = 'featured_entries'
            queryset = BlogEntry.objects.all()

        blog_entries = BlogEntry.objects.filter(status='published')
        manager['blog__featured_entries'] = blog_entries

        # accessing the value will return the model queryset
        assert manager['blog__featured_entries'] == blog_entries

    .. note::

        You should provide either the :py:attr:`queryset` or :py:attr:`model`
        attribute
    """

    serializer_class = ModelMultipleSerializer
    field_class = forms.ModelMultipleChoiceField

    def _setup_signals(self):
        pass


class FilePreference(BasePreferenceType):
    """
    A preference type that stores a a reference to a model.

    :Example:

    .. code-block:: python

        from django.core.files.uploadedfile import SimpleUploadedFile

        @registry.register
        class Logo(FilePreference):
            section = Section('blog')
            name = 'logo'

        logo = SimpleUploadedFile(
            "logo.png", b"file_content", content_type="image/png")
        manager['blog__logo'] = logo

        # accessing the value will return a FieldFile object, just as
        # django.db.models.FileField
        assert manager['blog__logo'].read() == b'file_content'

        manager['blog__logo'].delete()

    """

    field_class = forms.FileField
    serializer_class = FileSerializer
    default = None

    @property
    def serializer(self):
        """
        The serializer need additional data about the related preference
        to upload file to correct directory
        """
        return self.serializer_class(self)

    def get_field_kwargs(self):
        kwargs = super(FilePreference, self).get_field_kwargs()
        kwargs["required"] = self.get("required", False)
        return kwargs

    def get_upload_path(self):
        return os.path.join(
            preferences_settings.FILE_PREFERENCE_UPLOAD_DIR, self.identifier()
        )

    def get_file_storage(self):
        """
        Override this method if you want to use a custom storage
        """
        return default_storage

    def api_repr(self, value):
        if value:
            return value.url


class DurationPreference(BasePreferenceType):
    """
    A preference type that stores a timedelta.
    """

    field_class = forms.DurationField
    serializer = DurationSerializer

    def api_repr(self, value):
        return duration_string(value)


class DatePreference(BasePreferenceType):
    """
    A preference type that stores a date.
    """

    field_class = forms.DateField
    serializer = DateSerializer

    def api_repr(self, value):
        return value.isoformat()


class DateTimePreference(BasePreferenceType):
    """
    A preference type that stores a datetime.
    """

    field_class = forms.DateTimeField
    serializer = DateTimeSerializer

    def api_repr(self, value):
        return value.isoformat()


class TimePreference(BasePreferenceType):
    """
    A preference type that stores a time.
    """

    field_class = forms.TimeField
    serializer = TimeSerializer

    def api_repr(self, value):
        return value.isoformat()


class MultipleChoicePreference(ChoicePreference):
    """
    A preference type that stores multiple strings among a list of choices.

    :Example:

    .. code-block:: python

        @registry.register
        class FeaturedEntries(MultipleChoicePreference):
            section = Section('blog')
            name = 'featured_entries'
            choices = [
                ('c', 'Carrot'),
                ('t', 'Tomato'),
            ]

    .. note::

       Internally, the selected choices are stored as a string, separated by a
       separator. The separator defaults to ','. The way this is implemented still
       is sae also on keys that cotain the separator, but if in doubt, you can still
       set the :py:attr:`separator` to any other character.
    """

    widget = forms.CheckboxSelectMultiple
    field_class = forms.MultipleChoiceField
    serializer = MultipleSerializer

    def validate(self, value):
        for v in value:
            super().validate(v)



try:
    from django.urls import include, re_path
except ImportError:
    from django.conf.urls import include, url as re_path

from django.contrib.admin.views.decorators import staff_member_required
from . import views
from .registries import global_preferences_registry
from .forms import GlobalPreferenceForm

app_name = "dynamic_preferences"

urlpatterns = [
    re_path(
        r"^global/$",
        staff_member_required(
            views.PreferenceFormView.as_view(
                registry=global_preferences_registry, form_class=GlobalPreferenceForm
            )
        ),
        name="global",
    ),
    re_path(
        r"^global/(?P<section>[\w\ ]+)$",
        staff_member_required(
            views.PreferenceFormView.as_view(
                registry=global_preferences_registry, form_class=GlobalPreferenceForm
            )
        ),
        name="global.section",
    ),
    re_path(r"^user/", include("dynamic_preferences.users.urls")),
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="GlobalPreferenceModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                        auto_created=True,
                    ),
                ),
                (
                    "section",
                    models.CharField(
                        blank=True,
                        default=None,
                        null=True,
                        max_length=150,
                        db_index=True,
                    ),
                ),
                ("name", models.CharField(max_length=150, db_index=True)),
                ("raw_value", models.TextField(blank=True, null=True)),
            ],
            options={
                "verbose_name_plural": "global preferences",
                "verbose_name": "global preference",
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name="globalpreferencemodel",
            unique_together=set([("section", "name")]),
        ),
    ]



from django.core.exceptions import FieldDoesNotExist
from django.apps import apps

# import the logging library
import warnings
import logging
import collections
import persisting_theory

# Get an instance of a logger
logger = logging.getLogger(__name__)


#: The package where autodiscover will try to find preferences to register

from .managers import PreferencesManager
from .settings import preferences_settings
from .exceptions import NotFoundInRegistry
from .types import StringPreference
from .preferences import EMPTY_SECTION, Section


class MissingPreference(StringPreference):
    """
    Used as a fallback when the preference object is not found in registries
    This can happen for example when you delete a preference in the code,
    but don't remove the corresponding entries in database
    """

    pass


class PreferenceModelsRegistry(persisting_theory.Registry):
    """Store relationships beetween preferences model and preferences registry"""

    look_into = preferences_settings.REGISTRY_MODULE

    def register(self, preference_model, preference_registry):
        self[preference_model] = preference_registry
        preference_registry.preference_model = preference_model
        if not hasattr(preference_model, "registry"):
            setattr(preference_model, "registry", preference_registry)
        self.attach_manager(preference_model, preference_registry)

    def attach_manager(self, model, registry):
        if not hasattr(model, "instance"):
            return

        def instance_getter(self):
            return registry.manager(instance=self)

        getter = property(instance_getter)
        instance_class = model._meta.get_field("instance").remote_field.model
        setattr(instance_class, preferences_settings.MANAGER_ATTRIBUTE, getter)

    def get_by_preference(self, preference):
        return self[
            preference._meta.proxy_for_model
            if preference._meta.proxy
            else preference.__class__
        ]

    def get_by_instance(self, instance):
        """Return a preference registry using a model instance"""
        # we iterate through registered preference models in order to get the instance class
        # and check if instance is an instance of this class
        for model, registry in self.items():
            try:
                instance_class = model._meta.get_field("instance").remote_field.model
                if isinstance(instance, instance_class):
                    return registry

            except FieldDoesNotExist:  # global preferences
                pass
        return None


preference_models = PreferenceModelsRegistry()


class PreferenceRegistry(persisting_theory.Registry):

    """
    Registries are special dictionaries that are used by dynamic-preferences to register and access your preferences.
    dynamic-preferences has one registry per Preference type:

    - :py:const:`user_preferences`
    - :py:const:`site_preferences`
    - :py:const:`global_preferences`

    In order to register preferences automatically, you must call :py:func:`autodiscover` in your URLconf.

    """

    look_into = preferences_settings.REGISTRY_MODULE

    #: a name to identify the registry
    name = "preferences_registry"
    preference_model = None

    #: used to reverse urls for sections in form views/templates
    section_url_namespace = None

    def __init__(self, *args, **kwargs):
        super(PreferenceRegistry, self).__init__(*args, **kwargs)
        self.section_objects = collections.OrderedDict()

    def register(self, preference_class):
        """
        Store the given preference class in the registry.

        :param preference_class: a :py:class:`prefs.Preference` subclass
        """
        preference = preference_class(registry=self)
        self.section_objects[preference.section.name] = preference.section

        try:
            self[preference.section.name][preference.name] = preference

        except KeyError:
            self[preference.section.name] = collections.OrderedDict()
            self[preference.section.name][preference.name] = preference

        return preference_class

    def _fallback(self, section_name, pref_name):
        """
        Create a fallback preference object,
        This is used when you have model instances that do not match
        any registered preferences, see #41
        """
        message = (
            "Creating a fallback preference with "
            + 'section "{}" and name "{}".'
            + "This means you have preferences in your database that "
            + "don't match any registered preference. "
            + "If you want to delete these entries, please refer to the "
            + "documentation: https://django-dynamic-preferences.readthedocs.io/en/latest/lifecycle.html"
        )  # NOQA
        warnings.warn(message.format(section_name, pref_name))

        class Fallback(MissingPreference):
            section = Section(name=section_name) if section_name else None
            name = pref_name
            default = ""
            help_text = "Obsolete: missing in registry"

        return Fallback()

    def get(self, name, section=None, fallback=False):
        """
        Returns a previously registered preference

        :param section: The section name under which the preference is registered
        :type section: str.
        :param name: The name of the preference. You can use dotted notation 'section.name' if you want to avoid providing section param
        :type name: str.
        :param fallback: Should we return a dummy preference object instead of raising an error if no preference is found?
        :type name: bool.
        :return: a :py:class:`prefs.BasePreference` instance
        """
        # try dotted notation
        try:
            _section, name = name.split(preferences_settings.SECTION_KEY_SEPARATOR)
            return self[_section][name]

        except ValueError:
            pass

        # use standard params
        try:
            return self[section][name]

        except KeyError:
            if fallback:
                return self._fallback(section_name=section, pref_name=name)
            raise NotFoundInRegistry(
                "No such preference in {0} with section={1} and name={2}".format(
                    self.__class__.__name__, section, name
                )
            )

    def get_by_name(self, name):
        """Get a preference by name only (no section)"""
        for section in self.values():
            for preference in section.values():
                if preference.name == name:
                    return preference
        raise NotFoundInRegistry(
            "No such preference in {0} with name={1}".format(
                self.__class__.__name__, name
            )
        )

    def manager(self, **kwargs):
        """Return a preference manager that can be used to retrieve preference values"""
        return PreferencesManager(registry=self, model=self.preference_model, **kwargs)

    def sections(self):
        """
        :return: a list of apps with registered preferences
        :rtype: list
        """

        return self.keys()

    def preferences(self, section=None):
        """
        Return a list of all registered preferences
        or a list of preferences registered for a given section

        :param section: The section name under which the preference is registered
        :type section: str.
        :return: a list of :py:class:`prefs.BasePreference` instances
        """

        if section is None:
            return [self[section][name] for section in self for name in self[section]]
        else:
            return [self[section][name] for name in self[section]]


class PerInstancePreferenceRegistry(PreferenceRegistry):
    pass


class GlobalPreferenceRegistry(PreferenceRegistry):
    section_url_namespace = "dynamic_preferences:global.section"

    def populate(self, **kwargs):
        return self.models(**kwargs)


global_preferences_registry = GlobalPreferenceRegistry()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

import dynamic_preferences

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = dynamic_preferences.__version__

if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    sys.exit()

readme = open("README.rst").read()

setup(
    name="django-dynamic-preferences",
    version=version,
    description="""Dynamic global and instance settings for your django project""",
    long_description=readme,
    author="Agate Blue",
    author_email="me+github@agate.blue",
    url="https://github.com/agateblue/django-dynamic-preferences",
    packages=["dynamic_preferences"],
    include_package_data=True,
    install_requires=[
        "django>=3.2",
        "six",
        "persisting_theory==1.0",
    ],
    license="BSD",
    zip_safe=False,
    keywords="django-dynamic-preferences",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("dynamic_preferences", "0004_move_user_model"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="globalpreferencemodel",
            options={
                "verbose_name": "Global preference",
                "verbose_name_plural": "Global preferences",
            },
        ),
        migrations.AlterField(
            model_name="globalpreferencemodel",
            name="name",
            field=models.CharField(db_index=True, max_length=150, verbose_name="Name"),
        ),
        migrations.AlterField(
            model_name="globalpreferencemodel",
            name="raw_value",
            field=models.TextField(blank=True, null=True, verbose_name="Raw Value"),
        ),
    ]



from django.contrib import admin as django_admin
from django import forms

from ..settings import preferences_settings
from .. import admin
from .models import UserPreferenceModel
from .forms import UserSinglePreferenceForm


class UserPreferenceAdmin(admin.PerInstancePreferenceAdmin):
    search_fields = ["instance__username"] + admin.DynamicPreferenceAdmin.search_fields
    form = UserSinglePreferenceForm
    changelist_form = UserSinglePreferenceForm

    def get_queryset(self, request, *args, **kwargs):
        # Instanciate default prefs
        getattr(request.user, preferences_settings.MANAGER_ATTRIBUTE).all()
        return super(UserPreferenceAdmin, self).get_queryset(request, *args, **kwargs)


django_admin.site.register(UserPreferenceModel, UserPreferenceAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ("dynamic_preferences", "0002_auto_20150712_0332"),
    ]

    operations = [
        migrations.AlterField(
            model_name="globalpreferencemodel",
            name="name",
            field=models.CharField(max_length=150, db_index=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="globalpreferencemodel",
            name="section",
            field=models.CharField(
                max_length=150,
                null=True,
                default=None,
                db_index=True,
                blank=True,
                verbose_name="Section Name",
            ),
            preserve_default=True,
        ),
    ]
